package org.jvnet.hifaces.binding;

import javax.faces.context.FacesContext;
import javax.faces.el.EvaluationException;
import javax.faces.el.PropertyNotFoundException;
import javax.faces.el.ValueBinding;

public abstract class ParameterizedValueBinding<T, V> extends
		org.jvnet.hifaces.binding.ValueBinding<V> {

	private static final long serialVersionUID = 1L;

	private final ValueBinding target;

	private final ValueBinding[] arguments;

	public ParameterizedValueBinding(final javax.faces.el.ValueBinding target,
			final javax.faces.el.ValueBinding[] arguments) {
		super();
		this.target = target;
		this.arguments = arguments;
	}

	@SuppressWarnings("unchecked")
	public Object getValue(FacesContext facesContext)
			throws EvaluationException, PropertyNotFoundException {
		final T base = getBase(facesContext);
		final Object[] args = getArguments(facesContext);
		return get(base, args);
	}

	private T getBase(FacesContext facesContext) {
		return (T) target.getValue(facesContext);
	}

	@SuppressWarnings("unchecked")
	@Override
	public void setValue(FacesContext facesContext, Object value)
			throws EvaluationException, PropertyNotFoundException {
		final T base = getBase(facesContext);
		final Object[] args = getArguments(facesContext);
		set(base, args, (V) value);
	}

	@Override
	public boolean isReadOnly(FacesContext facesContext)
			throws EvaluationException, PropertyNotFoundException {
		return true;
	}

	private Object[] getArguments(FacesContext facesContext) {
		final Object[] args = new Object[arguments.length];
		for (int index = 0; index < arguments.length; index++) {
			args[index] = arguments[index].getValue(facesContext);
		}
		return args;
	}

	public abstract V get(T base, Object[] arguments);

	public void set(T base, Object[] arguments, V value) {
		throw new UnsupportedOperationException();
	}

	public abstract static class Single<T, U, V> extends
			ParameterizedValueBinding<T, V> {

		public Single(javax.faces.el.ValueBinding base,
				javax.faces.el.ValueBinding argument) {
			super(base, new ValueBinding[] { argument });
		}

		@SuppressWarnings("unchecked")
		public V get(T base, Object[] arguments) {
			return get(base, (U) arguments[0]);
		}

		@SuppressWarnings("unchecked")
		public void set(T base, Object[] arguments, V value) {
			set(base, value, (U) arguments[0]);
		}

		public abstract V get(T base, U argument);

		public void set(T base, V value, U argument) {
			throw new UnsupportedOperationException();
		}

	}
}
